/* $Id: cfg.h 422 2010-08-14 05:10:55Z oh2gve $
 * 
 * Copyright 2007, 2008, 2009, 2010 Tapio Aaltonen
 * 
 * 
 * This file is part of aprsg.
 *
 * aprsg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * aprsg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with aprsg.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * APRS is a registered trademark of APRS Software and Bob Bruninga, WB4APR.
*/

#ifndef CFG_H
#define CFG_H

// own stuff.
#include "filter.h"
#include "aprspacket.h"

// std stuff.
#include <vector>

// wx stuff.
#include <wx/string.h>
#include <wx/socket.h>


/// Module for reading and configuration.
/**
 * Defines structures for representing configuration data and one function to
 * read config from given file.
 */
namespace Cfg
{
	/// Types of connection.
	enum ConnectionType
	{   
		AX25PORT, SERIALKISSTNC, SERIALTNC
	};

	/// States of connection.
	enum ConnectionState
	{
		READONLY, WRITEONLY, READWRITE
	};
	
	/// Parameters of a serial port.
	struct SerialCfg
	{
		/// Serial port speed in bps.
		unsigned int speed;
		
		/// Serial port word length in bits.
		unsigned int charlength;
		
		/// Serial port parity: None, Odd or Even.
		char parity;
		
		/// Serial port stop bit count.
		unsigned int stopbits;
		
		/// HW flow control.
		bool rtscts;
	};
	
	/// Cfg of one Limiter.
	struct LimiterCfg
	{
		/// Holding time for packets going out through this connection.
		time_t holdtime;
		
		/// Amount for packets per time to accept.
		unsigned int amountOfPackets;
		
		/// Amount of time to count packets.
		time_t amountOfTime;
	};
	
	/// Cfg of one connection.
	struct ConnectionCfg
	{
		/// Name of the connection.
		wxString name;
		
		/// Type of the connection.
		ConnectionType type;
		/// State of the connection.
		ConnectionState state;

		/// Device name or file to open.
		wxString device;
		/// If type = serial, here are the port params.
		SerialCfg portcfg;
		/// Should the connection die upon error or not.
		bool dieOnError;

		/// Filter elements for this connection.
		std::vector<Filter> filterElements;
		/// Cfg for the Limiter of this connection.
		LimiterCfg limitercfg;
		/// Output path of the connection.
		std::vector<wxString> path;
	};
	
	/// Cfg of our uplink.
	struct UplinkCfg 
	{
		// Sekrit code.
		wxString passcode;
		
		/// APRS-IS server list.
		std::vector<wxIPV4address> servers;
		
		/// Original names of servers.
		std::vector<wxString> serverNames;

		/// State of the server connection.
		ConnectionState serverState;
		
		/// Amount of seconds to wait for incoming traffic.
		int noTrafficTimeout;
		
		/// Amount of seconds to wait between connection attempts.
		int retryInterval;
		
		/// Filter for server.
		wxString filter;
	};

	/// Logging setup.
	struct LogCfg
	{
		/// Path and filename of log file.
		wxString mainFileName;
		
		/// Path and filename of packet log file.
		wxString packetFileName;

		/// Host name or ip and port where to listen.
		wxIPV4address packetSrv;
		
		/// Packet log server status.
		bool serverNeeded;
	};
	
	/// Whole content of configuration file is stored like this.
	struct Configuration
	{
		/// Callsign of this igate.
		wxString callsign;
		
		/// Location of this igate.
		AprsPacket::Location location;

		/// Comment of our igate.
		wxString comment;

		/// Beacon packet sending interval in seconds.
		int beaconInterval;
		
		/// Uplink cfg.
		UplinkCfg uplink;
		
		/// Configurations of connections.
		std::vector<ConnectionCfg> connections;

		/// Amount of seconds to store packets for duplicate preventation.
		int duplicateTime;
		
		/// Logging params.
		LogCfg logcfg;
		
		/// Lock file directory.
		wxString lockdir;
		
		/// Pid file directory.
		wxString rundir;
		
		/// Object file name.
		wxString objectFile;
	};


	/// Read config file and return it's contents.
	/**
	 * Read contents of configuration file to config struct. Returns
	 * false, if required key or value for a key is not found, or
	 * parsing of a value fails. Extra keys are ignored without any
	 * warning.
	 */ 
	bool readCfg(const wxString& cfgFileName, Configuration& configuration);
	
} // namespace Cfg

	
#endif // CFG_H
