/* $Id: guard.cpp 411 2010-05-06 19:29:33Z oh2gve $
 * 
 * Copyright 2009, 2010 Tapio Aaltonen
 *
 * 
 * This file is part of aprsg.
 *
 * aprsg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * aprsg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with aprsg.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * APRS is a registered trademark of APRS Software and Bob Bruninga, WB4APR.
*/

#ifdef HAVE_CONFIG_H
#	ifdef __WXMSW__
#		include "../windows/config.h"
#	else
#		include <config.h>
#	endif
#endif

// own stuff.
#include "guard.h"
#include "packet.h"
#include "aprspacket.h"

// std stuff.
#include <string>
using std::string;
#include <list>
using std::list;
#include <map>
using std::map;
using std::make_pair;
#include <vector>
using std::vector;
#include <ctime>

// wx stuff.
#include <wx/log.h>



Guard::Guard(Connection* myconn, Cfg::ConnectionCfg const& mycfg, unsigned int const& mynum, unsigned int const& concount, DupeFilter& dupefilter,Packetlog& packetlog, string const& owncall):
myconn_(myconn), mycfg_(mycfg), mynum_(mynum+48),
dupefilter_(dupefilter), packetlog_(packetlog), owncall_(owncall), hold_(), clients_(), msgers_(),
packetCount_(0), timer_(time(NULL))
{
}



Guard::~Guard()
{
	while ( !hold_.empty() )
	{
		delete hold_.front()->aprsPacket;
		delete hold_.front();
		hold_.pop_front();
	}
}



Packet* Guard::givePacket(unsigned int const& timeout)
{
	AprsPacket* aprspacket = myconn_->readTimeout(timeout);
	if ( aprspacket )
	{
		// Create result object.
		Packet* result = new Packet;
		result->timestamp = wxDateTime::Now();
		result->input = mynum_;
		result->aprsPacket = aprspacket;
		result->thirdParty = NULL;
		result->output = '_';
		result->flag = '-';
			
		// Check if the packet is valid.
		string msg;
		if ( aprspacket->error(msg) )
		{
			// Print the packet.
			result->flag = 'e';
			packetlog_.write(result);
			return NULL;
		}
		
		// Try to get third-party packet from the packet.
		result->thirdParty = aprspacket->getCarriedThirdParty();
		while ( result->thirdParty )
		{
			// Update pointer to innermost packet.
			aprspacket = result->thirdParty;
			
			// Get next packet.
			AprsPacket* next3p = result->thirdParty->getCarriedThirdParty();
			
			// If we got someting, we move to it.
			if ( next3p )
			{
				delete result->thirdParty;
				result->thirdParty = next3p;
				aprspacket = next3p;
			}
			else
			{
				break;
			}
		}
		
		// Drop possible copy of this packet from our output hold.
		for ( list<Packet const*>::iterator i = hold_.begin(); i != hold_.end(); ++i )
		{
			if ( *aprspacket == *(*i)->aprsPacket )
			{
				delete (*i)->aprsPacket;
				delete *i;
				hold_.erase(i);
				break;
			}
		}
		
		// If this is a location packet from a messager, drop the messager from pass list of our tx filter.
		map<string, time_t>::iterator msgeriter = msgers_.find(aprspacket->getSource());
		if ( msgeriter != msgers_.end() )
		{
#ifdef __WXDEBUG__
			wxLogMessage(wxT(" debug(c") + wxString::Format(wxT("%c"), mynum_) + wxT("): packet from ") + wxString::FromAscii(aprspacket->getLabel().c_str()) + wxT(", he's on messager list."));
#endif
			// Check if packet contains location.
			AprsPacket::Location loc;
			if ( aprspacket->getLocation(loc) )
			{
				msgers_.erase(msgeriter);
#ifdef __WXDEBUG__
				wxLogMessage(wxT(" debug(c") + wxString::Format(wxT("%c"), mynum_) + wxT("): dropped messager ") + wxString::FromAscii(aprspacket->getLabel().c_str()) );
#endif
			}
#ifdef __WXDEBUG__
			else
			{
				wxLogMessage(wxT(" debug(c") + wxString::Format(wxT("%c"), mynum_) + wxT("): messager ") + wxString::FromAscii(aprspacket->getLabel().c_str()) + wxT(" was not dropped (not loc pkt)."));
			}
#endif
		}
		
		// Check if the packet source is in our client list.
		map<string, time_t>::iterator tmpClient = clients_.find(aprspacket->getLabel());
		
		// Check if the packet seen on RF passes our filter. Objects, items or stations seen only in third party format are not our clients.
		if ( !result->thirdParty && aprspacket->getSource() == aprspacket->getLabel() && passesFilter(aprspacket) )
		{
			// If this was a new client, make log entry.
			if ( tmpClient == clients_.end() )
			{
				wxLogMessage(wxT(" connection") + wxString::Format(wxT("%c"), mynum_) + wxT(": added messaging client ") + wxString::FromAscii(aprspacket->getLabel().c_str()) + wxT(""));
			}
			
			// Save the target into our client list.
			clients_[aprspacket->getLabel()] = time(NULL);

		}
		else if ( !result->thirdParty && !aprspacket->getMessage(msg, msg) )
		{
			// Target is now object or item, or doesn't pass our filter. Check if we used to gate messages to it.
			if ( tmpClient != clients_.end() )
			{
				// Yes we did, we need to stop doing that now.
				wxLogMessage(wxT(" connection") + wxString::Format(wxT("%c"), mynum_) + wxT(": dropped messaging client ") + wxString::FromAscii(aprspacket->getLabel().c_str()) + wxT(" (not reachable anymore)"));
				clients_.erase(tmpClient);
			}
		}
		
		// Check for duplicateness.
		if ( dupefilter_.isDupe(aprspacket) )
		{
			result->flag = 'd';
			packetlog_.write(result);
			return NULL;
		}
		
		// Check if the packet can be gated.
		if ( !aprspacket->canBeGated() )
		{
			result->flag = 'n';
			packetlog_.write(result);
			return NULL;
		}
		
		// Return the packet.
		return result;
	}
	
	return NULL;
}



bool Guard::passesFilter(AprsPacket const* packet)
{
	// Check if packet is a message packet.
	string message, destination;
	if ( packet->getMessage(message, destination) )
	{
		// Check if destination is our client.
		for ( map<string, time_t>::const_iterator i = clients_.begin(); i != clients_.end(); ++i )
		{
			if ( i->first == destination )
			{
				// Yes it is, save sender as messager.
				msgers_[packet->getSource()] = time(NULL);
				wxLogMessage(wxT(" connection") + wxString::Format(wxT("%c"), mynum_) + wxT(": added message sender ") + wxString::FromAscii(packet->getSource().c_str()) + wxT(" to temporary pass list"));
				
				// Passes.
				return true;
			}
		}
	}
	
	// Check if packet contains location.
	AprsPacket::Location loc;
	bool hasLocation = false;
	if ( packet->getLocation(loc) )
	{
		hasLocation = true;
	}
	
	// If this is a location packet of a messager, it passes.
	if ( hasLocation )
	{
		map<string, time_t>::iterator pos = msgers_.find(packet->getSource());
		if ( pos != msgers_.end() )
		{
			// Only one position packet per message one message packet is passed.
			wxLogMessage(wxT(" connection") + wxString::Format(wxT("%c"), mynum_) + wxT(": ") + wxString::FromAscii(packet->getSource().c_str()) + wxT(" is on temporary pass list, gating packet and dropping from the list"));
			msgers_.erase(pos);
			return true;
		}
	}
	
	// Check packet agains filter elements.
	for ( unsigned int i = 0; i < mycfg_.filterElements.size(); ++i )
	{
		if ( hasLocation )
		{
			if ( mycfg_.filterElements.at(i).accepts(packet->getLabel(), loc.lon, loc.lat) )
			{
				return true;
			}
		}
		else
		{
			if ( mycfg_.filterElements.at(i).accepts(packet->getLabel()) )
			{
				return true;
			}
		}      
	}

	// Every element said no-go.
	return false;
}



void Guard::sendPacket(Packet* packet)
{
	// Hold if necessary, send straigt away if not.
	if ( mycfg_.limitercfg.holdtime > 0 )
	{
		hold_.push_back(packet);
	}
	else
	{
		transmit(packet);
		delete packet->aprsPacket;
		delete packet;
	}
}



void Guard::kick()
{
	if ( !hold_.empty() )
	{
		// Check timestamp of the oldest packet.
		if ( time(0) - hold_.front()->aprsPacket->getTimestamp() > mycfg_.limitercfg.holdtime )
		{
			// Okay, it's now old enough to be transmitted.
			transmit(hold_.front());
			
			// Packet processed.
			delete hold_.front()->aprsPacket;
			delete hold_.front();
			hold_.pop_front();
		}
	}
	
	// Check if limit time cycle has completed.
	if ( myconn_->outputEnabled() && time(NULL) - timer_ >= mycfg_.limitercfg.amountOfTime )
	{
		// Cycle is complete, reset packet counter and start new cycle.
		packetCount_ = 0;
		timer_ = time(NULL);
	}
	
	// Flush client list.
	for ( map<string, time_t>::iterator i = clients_.begin(); i != clients_.end(); ++i )
	{
		// Check if last info from the client is older than 30 mins.
		if ( time(NULL) - i->second > 60*30 )
		{
			wxLogMessage(wxT(" connection") + wxString::Format(wxT("%c"), mynum_) + wxT(": dropped messaging client ") + wxString::FromAscii(i->first.c_str()) + wxT(" (no pkt in last 30 min)"));
			clients_.erase(i);
			// This function is run often enough to remove just one client at a time.
			break;
		}
	}

	// Flush messager list.
	for ( map<string, time_t>::iterator i = msgers_.begin(); i != msgers_.end(); ++i )
	{
		// Check if last message from the messager is older than 30 mins.
		if ( time(NULL) - i->second > 60*30 )
		{
			wxLogMessage(wxT(" connection") + wxString::Format(wxT("%c"), mynum_) + wxT(": dropped message source ") + wxString::FromAscii(i->first.c_str()) + wxT(" from temporary pass list (no loc pkt in last 30 min)"));
			msgers_.erase(i);
			break;
		}
	}
}



void Guard::transmit(Packet const* input)
{
	// First we turn the packet into third party format.
	string packet ( input->aprsPacket->getAsThirdParty( owncall_, input->inputName ) );
	
	// Next we get path we're going to use for the carrying frame.
	vector<string> outpath;
	for ( unsigned int i = 0; i < mycfg_.path.size(); ++i )
	{
		outpath.push_back( string(mycfg_.path.at(i).ToAscii()) );
	}
	
	// Create ourselves a packet with the outgoing packet as body.
	AprsPacket* aprspacket = new AprsPacket( owncall_, string(OWN_TARGET), outpath, packet, true );
	string msg;
	if ( aprspacket->error(msg) )
	{
		wxLogMessage(wxT(" connection") + wxString::Format(wxT("%c"), mynum_) + wxT(": failed to construct outgoing packet (") + wxString::FromAscii(msg.c_str()) + wxT(")"));
		delete aprspacket;
		return;
	}
			
	// Now we create a log copy of the packet.
	Packet* logPkt = new Packet(*input);
	logPkt->timestamp = wxDateTime::Now();
	logPkt->aprsPacket = new AprsPacket(*aprspacket);
	logPkt->input = '_';
	logPkt->output = mynum_;
	
	// Check if we're below maxrate.
	if ( packetCount_ < mycfg_.limitercfg.amountOfPackets )
	{
		// Yes we are, let's send the packet.
		myconn_->write(aprspacket);
		
		// Increase count.
		packetCount_++;
	}
	else
	{
		// We're above maxrate, let's just delete the packet.
		delete aprspacket;
		
		// Mark the packet limited.
		logPkt->flag = 'l';
	}
	
	// Save packet into log. If transmission was instant, raise the log flag to keep this appearing before the original received packet.
	if ( mycfg_.limitercfg.holdtime > 0 )
	{
		packetlog_.write(logPkt);
	}
	else
	{
		packetlog_.write(logPkt, true);
	}
}
